/*
 
  This program sets the 'memory clock' of Cirrus 542x cards.
  The first three values could be set by utility programs that
  came with my card (AVGA3), but much higher values seem to work (on my card
  at least). It may be that better and more recent Cirrus cards use a higher
  value as boot-up default. It should depend on DRAM speed, but it seems
  to be more dependant on the card logic.
  
  Note: it seems that the memory clock is linked with the VLB speed. The
  default memory clock of my card seems too high when running at 25MHz!

  Using a higher memory clock gives a very significant performance improvement;
  with high dot clock modes (like 640x480x16M or 1150x900x256) performance can
  be more than twice that of the standard 50 MHz clock. This goes for both
  (VLB) framebuffer access and accelerated features (bitblt). This also helps
  XFree86 server performance, but only if the XFree86 Cirrus driver doesn't
  set the memory clock register (it should work for XFree86 1.3 and 2.0).
  Use at your own risk!

  Note that the 'dot clock' is something entirely different. There does not
  seem to be much correlation between the two (i.e. if a high dot clock gives
  screen problems, using a high memory clock is not likely to fix it, other
  than improving speed).

*/


#include <stdlib.h>
#include <unistd.h>
#include "libvga.h"	/* For port I/O macros. */

void main() {
	iopl(3);
	outb(0x3c4, 0x1f);

	/*
		Tested on VLB AVGA3 CL-GD5426 on 40MHz VLB
		0x1c		50M (boot-up default)
		0x21		59M
		0x22		62M
		0x23 OK
		0x24 OK
		0x25 OK
		0x26 OK
		0x27 occasional loose pixels
		0x28 occasional problems
		0x29 some problems
		0x2A 16M/textmode problems
		0x2c 256/32k color problems
	*/
	outb(0x3c5, 0x26);

	exit(0);
}
