/*
**  tseng3.c - get ET4000 graphics register values
**  Copyright (C) 1993  Tommy Frandsen, Harm Hanemaayer, Hartmut Schirmer
**
**  This program is free software;
**
**  Permission is granted to any individual or institution to use, copy,
**  or redistribute this executable so long as it is not modified and
**  that it is not sold for profit.
**
**  LIKE ANYTHING ELSE THAT'S FREE, TSENG3 IS PROVIDED AS IS AND
**  COMEs WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED.
**  IN NO EVENT WILL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES
**  RESULTING FROM THE USE OF THIS SOFTWARE.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <stdarg.h>

/* VGA index register ports */
#define CRT_I   0x3D4   /* CRT Controller Index (mono: 0x3B4) */
#define ATT_IW  0x3C0   /* Attribute Controller Index & Data Write Register */
#define GRA_I   0x3CE   /* Graphics Controller Index */
#define SEQ_I   0x3C4   /* Sequencer Index */
#define PEL_IW  0x3C8   /* PEL Write Index */

/* VGA data register ports */
#define CRT_D   0x3D5   /* CRT Controller Data Register (mono: 0x3B5) */
#define ATT_R   0x3C1   /* Attribute Controller Data Read Register */
#define GRA_D   0x3CF   /* Graphics Controller Data Register */
#define SEQ_D   0x3C5   /* Sequencer Data Register */
#define MIS_R   0x3CC   /* Misc Output Read Register */
#define MIS_W   0x3C2   /* Misc Output Write Register */
#define IS1_R   0x3DA   /* Input Status Register 1 (mono: 0x3BA) */
#define PEL_D   0x3C9   /* PEL Data Register */

/* VGA indexes max counts */
#define CRT_C   24              /* 24 CRT Controller Registers */
#define ATT_C   21              /* 21 Attribute Controller Registers */
#define GRA_C   9               /* 9  Graphics Controller Registers */
#define SEQ_C   5               /* 5  Sequencer Registers */
#define MIS_C   1               /* 1  Misc Output Register */
#define EXT_C   11              /* 11 SVGA Extended Registers */

/* VGA registers saving indexes */
#define CRT     0               /* CRT Controller Registers start */
#define ATT     CRT+CRT_C       /* Attribute Controller Registers start */
#define GRA     ATT+ATT_C       /* Graphics Controller Registers start */
#define SEQ     GRA+GRA_C       /* Sequencer Registers */
#define MIS     SEQ+SEQ_C       /* General Registers */
#define EXT     MIS+MIS_C       /* SVGA Extended Registers */


unsigned char vga_regs[71];
static int do_all = 0;
FILE *output = NULL;

void port_out(unsigned char value, unsigned short port)
{
	asm mov dx,port
	asm mov al,value
	asm out dx,al
}


unsigned char port_in(unsigned short port)
{
	asm mov dx,port
	asm in al,dx
    return (_AL);
}

int dactype(void)
{
	union REGS cpu_regs;

	cpu_regs.x.ax=0x10F1;
	cpu_regs.x.bx=0x2EFF;

	int86(0x10, &cpu_regs, &cpu_regs);
	if (cpu_regs.x.ax != 0x0010)
		return (-1);
	else
		return (cpu_regs.h.bl);
}

void puts2(char *s) {
  if (output != NULL)
    fprintf(output, "%s\n", s);
  while (*s != '\0') {
    if (*s == '\n') putch('\r');
    putch (*(s++));
  }
  putch('\r'); putch('\n');
}

int printf2(char *fmt, ...)
{
  va_list  argptr;
  char str[140], *s;
  int cnt;

  va_start( argptr, fmt );

  if (output != NULL)
    vfprintf(output, fmt, argptr);
  cnt = vsprintf( str, fmt, argptr );
  s = str;
  while (*s) {
    switch (*s) {
      case '\n' : putch('\r');
		  break;
      case '\t' : cprintf("%*s", 8-((wherex()-1)&7), "");
		  ++s;
		  continue;
    }
    putch(*(s++));
  }
  va_end( argptr );
  return( cnt );
}

void get_dac(void)
{
    int dac;

    dac = dactype();
    switch (dac)
    {
	case -1:
		printf2 ("/* Dac detection BIOS call returned an error */\n");
		break;
	case 0:
		printf2 ("/* Standard VGA dac detected */\n");
		printf2 ("#define DAC_TYPE 0\n");
		break;
	case 1:
		printf2 ("/* Sierra SC1148x HiColor dac detected */\n");
		printf2 ("#define DAC_TYPE 1\n");
		break;
	case 2:
		printf2 ("/* Diamond Speedstar 24 24bit dac or Sierra Mark2/Mark3 dac detected */\n");
		break;
	case 3:
		printf2 ("/* AT&T ATT20c491/2 15/16/24 bit dac detected */\n");
		printf2 ("#define DAC_TYPE 9\n");
		break;
	case 4:
		printf2 ("/* AcuMos ADAC1 15/16/24 bit dac detected */\n");
		break;
	case 8:
		printf2 ("/* Music 15/16/24 bit dac (AT&T compatible) detected */\n");
		printf2 ("#define DAC_TYPE 9\n");
		break;
	default:
		printf2 ("/* Unknown HiColor dac (%d) detected */\n", dac);
		break;
    }
}

#define TickTime (1.0/1193182.0)

#define init_timer() do { \
  asm mov al, 034h;       \
  asm out 043h, al;       \
  asm xor al, al;         \
  asm out 040h, al;       \
  asm out 040h, al;       \
} while(0)

short ReadTimer(void)
{
  asm  mov   al, 4;
  asm  out   43h, al;
  asm  in    al, 40h;
  asm  mov   bl, al;
  asm  in    al, 40h;
  asm  mov   ah, al;
  asm  mov   al, bl;
  asm  not   ax;
  return _AX;
}

#define CRTC_BASE (*((short *)MK_FP(0x40, 0x63)))

#define _wait_(ID,VAL)  \
w_##ID##1:              \
  asm in    al, dx;     \
  asm test  al, VAL;    \
  asm jne   w_##ID##1;  \
w_##ID##2:              \
  asm in    al, dx;     \
  asm test  al, VAL;    \
  asm je    w_##ID##2

#define wait_horizontal(ID,port)  do { _DX=(port); _wait_(ID##ch,8);} while(0)
#define wait_vertical(ID,port)    do { _DX=(port); _wait_(ID##cv,1);} while(0)

int interlaced(void) {
  return (vga_regs[EXT+3] & 0x80) ? 1 : 0;
}

int calc_vtotal(void) {
  int total;

  total = (vga_regs[EXT+3]&2) ? 1024 : 0;
  switch (vga_regs[CRT+7]&0x21) {
    case 0x01 : total += 256; break;
    case 0x20 : total += 512; break;
    case 0x21 : total += 768; break;
  }
  total += vga_regs[CRT+6];
  return total + 2;
}

double measure_horizontal(int mono)
{
  short start, stop;
  int   count;
  long  sum = 0;
  long  diff;

  if (mono) mono = 0x3ba; else mono = 0x3da;
  for (count = 0; count < 5; ) {
    disable();
    init_timer();
    wait_horizontal(mv0,mono);
    wait_vertical(mv0,mono);
    start = ReadTimer();
    wait_vertical(mv1,mono);
    wait_vertical(mv2,mono);
    wait_vertical(mv3,mono);
    wait_vertical(mv4,mono);
    wait_vertical(mv5,mono);
    wait_vertical(mv6,mono);
    wait_vertical(mv7,mono);
    wait_vertical(mv8,mono);
    wait_vertical(mv9,mono);
    wait_vertical(mv10,mono);
    wait_vertical(mv11,mono);
    wait_vertical(mv12,mono);
    wait_vertical(mv13,mono);
    wait_vertical(mv14,mono);
    wait_vertical(mv15,mono);
    stop = ReadTimer();
    enable();
    diff = stop-start;
    if (diff < 0) diff += 65536L;
    sum += diff;
    ++count;
    if (kbhit()) break;
  }
  return 15/(TickTime*((double)sum) / ((double) count));
}


void one_mode(int mode, int hico, int funny_card)
{
    union REGS cpu_regs;
    int i, bioserr=0, skip = 0;
    int x=0, y, col, size;
    char resol[100], colors[100], *desc;
    char   *verttxt;
    double vertical, horizontal;

    switch (mode) {
	case 0x13: x =  320; y =  200; col = 256; break;
	case 0x25: x =  640; y =  480; col =  16; break;
	case 0x29: x =  800; y =  600; col =  16; break;
	case 0x2d: x =  640; y =  350; col = 256; break;
	case 0x2e: x =  640; y =  480; col = 256; break;
	case 0x2f: x =  640; y =  400; col = 256; break;
	case 0x30: x =  800; y =  600; col = 256; break;
	case 0x37: x = 1024; y =  768; col =  16; break;
	case 0x38: x = 1024; y =  768; col = 256; break;
	case 0x3d: x = 1280; y = 1024; col =  16; break;
	default  : return;
    }
    if (hico != 0 && col != 256) return;
    size = (((unsigned long) x) * ((unsigned long) y)) >> 10;
    if (col == 16) size >>= 2;

    switch (hico) {
	case 15: size*=2; strcpy(colors, "32K"); desc = "HiColor ";   break;
	case 16: size*=2; strcpy(colors, "64K"); desc = "HiColor ";   break;
	case 24: size*=3; strcpy(colors, "16M"); desc = "TrueColor "; break;
	default: sprintf(colors, "%d", col); desc = ""; break;
    }
    sprintf(resol, "%dx%dx%s", x, y, colors);

    cprintf ("\nͻ\r\n");
    cprintf (  " WARNING: Improper use of this program may destroy your monitor \r\n");
    cprintf (  "ͼ\r\n\n");
    cprintf("BIOS mode 0x%2X : %s\r\n", mode, resol);
    cprintf("\n");
    if (!do_all) {
      int ch;
      cprintf("Type SPACE to skip mode, ESC to cancel program, RETURN to continue ... ");
      do {
	ch = getch();
	if (ch == 27) exit(1);
	skip = (ch == ' ');
	do_all = (ch == 'A');
      } while (!skip && !do_all && ch != '\r');
    }

    if (!skip) {
      switch (hico)
      {
	  case 15:
	  case 16:
	      cpu_regs.x.ax=0x10F0;
	      cpu_regs.h.bl=mode;
	      break;
	  case 24:
	      switch (funny_card)
	      {
		  case 1:
		      cpu_regs.x.ax=0x10E0;
		      cpu_regs.h.bl=0x2e;
		      break;
		  case 2:
		      cpu_regs.x.ax=0x10f0;
		      cpu_regs.h.bl=0x3e;
		      break;
		  case 0:
		  default:
		      cpu_regs.x.ax=0x10F0;
		      cpu_regs.h.bh=mode;
		      cpu_regs.h.bl=0xFF;
		      break;
	      }
	      break;
	  case 0:
	  default:
	      cpu_regs.h.ah = 0x00;
	      cpu_regs.h.al = mode;
	      break;
	  }

	  int86(0x10, &cpu_regs, &cpu_regs);
	  bioserr = cpu_regs.x.ax;

      /* get VGA register values */
      for (i = 0; i < CRT_C; i++) {
	  port_out(i, CRT_I);
	  vga_regs[CRT+i] = port_in(CRT_D);
      }
      for (i = 0; i < ATT_C; i++) {
	  if (col == 16 && i < 16)
	    vga_regs[ATT+i] = i;
	  else {
	    port_in(IS1_R);
	    port_out(i, ATT_IW);
	    vga_regs[ATT+i] = port_in(ATT_R);
	  }
      }
      for (i = 0; i < GRA_C; i++) {
	  if (i==1 && col == 16)
	    vga_regs[GRA+i] = 0x0F;
	  else {
	    port_out(i, GRA_I);
	    vga_regs[GRA+i] = port_in(GRA_D);
	  }
      }
      for (i = 0; i < SEQ_C; i++) {
	  port_out(i, SEQ_I);
	  vga_regs[SEQ+i] = port_in(SEQ_D);
      }
      vga_regs[MIS] = port_in(MIS_R);

      /* get extended CRT registers */
      for (i = 0; i < 6; i++) {
	   port_out(0x32+i, CRT_I);
	   vga_regs[EXT+i] = port_in(CRT_D);
      }
      port_out(0x3f	, CRT_I);
      vga_regs[EXT+6] = port_in(CRT_D);

      /* get extended sequencer register */
      port_out(7, SEQ_I);
      vga_regs[EXT+7] = port_in(SEQ_D);

      /* get some other ET4000 specific registers */
      vga_regs[EXT+8] = port_in(0x3c3);
      vga_regs[EXT+9] = port_in(0x3cd);

      /* get extended attribute register */
      port_in(IS1_R);   /* reset flip flop */
      port_out(0x16, ATT_IW);
      vga_regs[EXT+10] = port_in(ATT_R);

      /* Measure video timing */
      horizontal = measure_horizontal(CRTC_BASE < 0x3d4);

      cpu_regs.h.ah = 0x00;
      cpu_regs.h.al = 0x03;
      int86(0x10, &cpu_regs, &cpu_regs);
    }

    printf2("\n/* ET4000 %sBIOS mode 0x%02X -- %s", desc, mode, resol);
    if ( skip ||
	 ( hico && (bioserr != 0x0010)) ||
	 (!hico && (bioserr == mode))    ) {
	printf2(" : NOT SUPPORTED */\n");
	if (size <= 1024)
	  printf2("#define g%s_regs DISABLE_MODE\n", resol);
	return;
    }
    printf2(" */\n");

    if (hico == 16) {
      printf2("#define g%s_regs g%dx%dx32K_regs\n", resol, x, y);
      return;
    }

    verttxt = interlaced() ? " (interlaced)" : "";
    vertical   = horizontal / calc_vtotal();
    printf2("/* Video timing:\tVertical frequency   : %4.1fHz%s\n",
						vertical, verttxt);
    printf2("\t\t\tHorizontal frequency : %4.1fKHz  */\n", horizontal/1000.0);
    if (mode == 0x3D)
      printf2("#define g1280x1024x16_regs DISABLE_MODE\n/* Doesn't work ...\n");
    printf2("static unsigned char g%s_regs[71] = {\n  ", resol);
    for (i = 0; i < 12; i++)
	printf2("0x%02X,",vga_regs[CRT+i]);
    printf2("\n  ");
    for (i = 12; i < CRT_C; i++)
	printf2("0x%02X,",vga_regs[CRT+i]);
    printf2("\n  ");
    for (i = 0; i < 12; i++)
	printf2("0x%02X,",vga_regs[ATT+i]);
    printf2("\n  ");
    for (i = 12; i < ATT_C; i++)
	printf2("0x%02X,",vga_regs[ATT+i]);
    printf2("\n  ");
    for (i = 0; i < GRA_C; i++)
	printf2("0x%02X,",vga_regs[GRA+i]);
    printf2("\n  ");
    for (i = 0; i < SEQ_C; i++)
	printf2("0x%02X,",vga_regs[SEQ+i]);
    printf2("\n  ");
    printf2("0x%02X,",vga_regs[MIS]);
    printf2("\n  ");
    for (i = 0; i < EXT_C-1; i++)
	printf2("0x%02X,",vga_regs[EXT+i]);
    printf2("0x%02X",vga_regs[EXT+EXT_C-1]);
    printf2("\n};\n");
    if (mode == 0x3D)
      puts2("*/");
}

void usage(void) {
  cputs("tseng3 <output_file> [SS24]");
  exit(1);
}

void main(int argc, char* argv[])
{
    int funny_card = 0;

    if (argc >= 2)
      output = fopen(argv[1], "w");

    if (output != NULL)
      fputs("/*\n   ( File generated by tseng3.exe )\n\n", output);
    puts2("tseng3 v1.0, Copyright (C) 1993  Tommy Frandsen, Harm Hanemaayer");
    puts2("and Hartmut Schirmer\n");
    puts2("Permission is granted to any individual or institution to use, copy, or");
    puts2("redistribute this executable so long as it is not modified and that it is");
    puts2("not sold for profit.\n");
    puts2("LIKE ANYTHING ELSE THAT'S FREE, TSENG3 IS PROVIDED AS IS AND COMES WITH");
    puts2("NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED. IN NO EVENT WILL");
    puts2("THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES RESULTING FROM THE USE OF");
    puts2("THIS SOFTWARE.");
    if (output != NULL)
      fputs("*/\n", output);
    puts2("");

    if (argc >= 2 && output == NULL)
	usage();

    if (argc>=3)
	funny_card = (strcmp("SS24", argv[2]) == 0) ? 1 : 2;

    get_dac();

    one_mode(0x13, 15, funny_card);  /* 320x200x32K */
    one_mode(0x13, 16, funny_card);  /* 320x200x64K */
    one_mode(0x13, 24, funny_card);  /* 320x200x16M */
    puts2("\n/* --- ET4000 specific modes */\n#ifdef _DYNAMIC_ONLY_");
    one_mode(0x2d,  0, funny_card);  /* 640x350x256 */
    one_mode(0x2d, 15, funny_card);  /* 640x350x32K */
    one_mode(0x2d, 16, funny_card);  /* 640x350x64K */
    one_mode(0x2d, 24, funny_card);  /* 640x350x16M */
    one_mode(0x2f,  0, funny_card);  /* 640x400x256 */
    one_mode(0x2f, 15, funny_card);  /* 640x400x32K */
    one_mode(0x2f, 16, funny_card);  /* 640x400x64K */
    one_mode(0x2f, 24, funny_card);  /* 640x400x16M */
    puts2("\n#endif /* defined(_DYNAMIC_ONLY_ALL_) */\n");
    one_mode(0x2e,  0, funny_card);  /* 640x480x256 */
    one_mode(0x2e, 15, funny_card);  /* 640x480x32K */
    one_mode(0x2e, 16, funny_card);  /* 640x480x64K */
    one_mode(0x2e, 24, funny_card);  /* 640x480x16M */
    one_mode(0x29,  0, funny_card);  /* 800x600x16 */
    one_mode(0x30,  0, funny_card);  /* 800x600x256 */
    one_mode(0x30, 15, funny_card);  /* 800x600x32K */
    one_mode(0x30, 16, funny_card);  /* 800x600x64K */
    one_mode(0x30, 24, funny_card);  /* 800x600x16M */
    one_mode(0x37,  0, funny_card);  /* 1024x768x16 */
    one_mode(0x38,  0, funny_card);  /* 1024x768x256 */
    one_mode(0x38, 15, funny_card);  /* 1024x768x32K */
    one_mode(0x38, 16, funny_card);  /* 1024x768x64K */
    one_mode(0x38, 24, funny_card);  /* 1024x768x16M */
    one_mode(0x3D,  0, funny_card);  /* 1280x1024x16 */
}

